/**
 * Field plugin base.
 *
 * A base class for field plugins.
 */
export class FieldPluginBase {

  constructor() {
    this.originalValue = null;
  }

  init() {}

  /**
   * Supported widgets.
   *
   * @type {[]}
   *   An array of widget ID's this plugin supports.
   */
  supportedWidgets = [];

  /**
   * Edit.
   *
   * @param EditableElement
   *   The element being edited.
   */
  edit(EditableElement) {
    Drupal.EditPlus.CurrentlyEditingElement = EditableElement;
    this.replacePageElementWithFormItem(EditableElement);
    this.focus(EditableElement);
    this.revealAncillary(EditableElement);
  }

  /**
   * Focus.
   *
   * Focus on the field after it has been clicked.
   *
   * @param EditableElement
   *   The element being edited.
   */
  focus(EditableElement) {
    // Prepare for the eventual de-focusing the form item. It would be easy to
    // add a formItem.onblur when the form item looses focus, but losing focus
    // on the form item doesn't necessarily mean we are done with it. e.g.
    // changing a text format in the sidebar. Let's keep track of what is
    // currently focused so that we can check if one of its ancillary elements
    // was clicked and remain focused.
    EditableElement.getFormItemWrapper().classList.add('edit-plus-focused');
    document.addEventListener('mousedown', this.watchFocus, true);

    this.focusElement(EditableElement)
  }

  /**
   * Focus element.
   *
   * This should be overridden by field plugins to provide field specific
   * instructions on how to give the element focus.
   *
   * @param EditableElement
   *   The element being edited.
   *
   * @param event
   */
  focusElement(EditableElement, event) {}

  /**
   * Watch focus.
   *
   * Once an element is being edited, monitor clicks to determine if the click
   * should cause the form item to lose focus. Field plugins can provide selectors
   * that do not make the form item loose focus when clicked. e.g. things in the
   * sidebar like field formatters etc.
   *
   * @param e
   *   The mousedown event.
   */
  watchFocus(e) {
    // Check that the clicked element is something that should cause the form
    // item to lose focus.
    const plugin = Drupal.EditPlus.CurrentlyEditingElement.plugin;
    const getCurrentlyFocusedSelectors = plugin.getFocusedSelectors();
    let remainFocused = false;
    for (let i = 0; i < getCurrentlyFocusedSelectors.length; i++) {
      const ignoredClick = e.target.closest(getCurrentlyFocusedSelectors[i]);
      if (ignoredClick !== null) {
        remainFocused = true;
        break;
      }
    }
    if (!remainFocused) {
      plugin.blur();
      document.querySelectorAll('.edit-plus-focused').forEach(element => {
        element.classList.remove('edit-plus-focused');
      });
    }
  }

  /**
   * Blur.
   *
   * Remove focus from the focused form item.
   */
  async blur() {
    const EditableElement = Drupal.EditPlus.CurrentlyEditingElement;
    document.removeEventListener('mousedown', this.watchFocus, true);
    if (Drupal.EditPlus.CurrentlyEditingElement.sizedPlaceholderHeightObserver) {
      Drupal.EditPlus.CurrentlyEditingElement.sizedPlaceholderHeightObserver.disconnect();
    }
    Drupal.EditPlus.CurrentlyEditingElement = null;
    this.hideAncillary(EditableElement);
    return this.blurElement(EditableElement);
  }

  /**
   * Blur element.
   *
   * This should be overridden by field plugins to provide field specific
   * instructions on how to remove focus from the element. Since detecting if
   * the field has changes occurs at this point, after editing, it's recommended
   * for field plugins to call this.replaceFormItemWithPageElement(EditableElement)
   * when appropriate.
   *
   * @param EditableElement
   *   The element being edited.
   *
   * @returns {Promise<void>}
   *   Getting the form inputs needs to be async since we may need to wait on
   *   a CKEditor to be attached before getting changes.
   */
  blurElement(EditableElement) {
    return Promise.resolve();
  }

  /**
   * Reveal ancillary.
   *
   * Reveal things in the sidebar that pertain to this field e.g. formatter
   *
   * @param EditableElement
   *   The element being edited.
   */
  revealAncillary(EditableElement) {}

  /**
   * Hide ancillary.
   *
   * Reveal things in the sidebar that pertain to this field e.g. formatter
   *
   * @param EditableElement
   *   The current form item.
   */
  hideAncillary(EditableElement) {
    EditableElement.getFormItemWrapper().classList.add('edit-plus-hidden');
  }

  /**
   * Replace page element with form item.
   *
   * Visually moves the hidden form item from the form in the sidebar over the
   * editable element that was clicked in the main page area.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replacePageElementWithFormItem(EditableElement) {
    const pageElement = EditableElement.getPageElementHandle();
    const formItem = EditableElement.getFormItemHandle();

    // Add a placeholder element that the form item will hover over.
    const sizedPlaceholderId = EditableElement.getSizedPlaceholderId()
    const sizedPlaceholder = document.createElement('div');
    sizedPlaceholder.id = sizedPlaceholderId;
    sizedPlaceholder.classList.add('sized-placeholder');
    pageElement.parentElement.insertBefore(sizedPlaceholder, pageElement);

    this.positionFormItemOverPageElement();
    document.addEventListener('scroll', this.positionFormItemOverPageElement);
    window.addEventListener('resize', this.positionFormItemOverPageElement);

    // Ensure the sizedPlaceholder height stays in sync with the formItem height.
    // As items are added or removed from the form item via ajax, the sizedPlaceholder
    // show grow or shrink accordingly.
    Drupal.EditPlus.CurrentlyEditingElement.sizedPlaceholderHeightObserver = new ResizeObserver(entries => {
      const sizedPlaceholder = document.querySelector('#' + sizedPlaceholderId);
      if (sizedPlaceholder) {
        for (let entry of entries) {
          sizedPlaceholder.style.height = entry.contentRect.height + 'px';
        }
      }
    });
    Drupal.EditPlus.CurrentlyEditingElement.sizedPlaceholderHeightObserver.observe(formItem);

    pageElement.classList.add('edit-plus-hidden');
    formItem.classList.add('edit-plus-editing');
    formItem.classList.remove('edit-plus-hidden');
  }

  /**
   * Replace form item with page element.
   *
   * Moves a form item back to the form in the sidebar from the main page area.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replaceFormItemWithPageElement(EditableElement) {
    document.removeEventListener('scroll', this.positionFormItemOverPageElement);
    window.removeEventListener('resize', this.positionFormItemOverPageElement);

    const pageElement = EditableElement.getPageElementHandle();
    const formItem = EditableElement.getFormItemHandle();

    formItem.removeAttribute('style');
    formItem.classList.remove('edit-plus-editing');
    const sizedPlaceholder = document.querySelector('#' + EditableElement.getSizedPlaceholderId());
    if (sizedPlaceholder) {
      sizedPlaceholder.remove();
    }

    formItem.classList.add('edit-plus-hidden');
    pageElement.classList.remove('edit-plus-hidden');
  }

  /**
   * Position form item over page element.
   *
   * The form item is position: fixed over the placeholder. The
   * placeholder not only pushes content further down the page as if the
   * editable element is in place there, but it also gives us the dimensions that
   * the editable element should have.
   */
  positionFormItemOverPageElement() {
    const EditableElement = Drupal.EditPlus.CurrentlyEditingElement;
    const formItem = EditableElement.getFormItemHandle();
    const popOutWidth = EditableElement.getFormItemWrapper().dataset.popOutWidth ?? 200;
    const rectangle = document.querySelector('#' + EditableElement.getSizedPlaceholderId()).getBoundingClientRect();

    formItem.style.position = 'fixed';

    if (popOutWidth < rectangle.width) {
      // The form item will fit inline.
      formItem.style.left = rectangle.left + 'px';
      formItem.style.top = rectangle.top + 'px';
      formItem.style.width = rectangle.width + 'px';
      formItem.classList.remove('form-item-popout');
    } else {
      // The form item won't fit in there, make a pop out.
      formItem.style.top = rectangle.top + 'px';
      const width = (parseInt(popOutWidth) + 20);
      // Center the popup over the sizedPlaceholder.
      let left = (rectangle.left + (rectangle.width / 2)) - (popOutWidth / 2);
      // Ensure we aren't hanging over the left edge.
      const leftEdge = Drupal.displace.calculateOffset('left');
      left = left > leftEdge ? left : leftEdge;
      // Ensure we aren't hanging over the right edge.
      const rightEdge = document.documentElement.clientWidth - Drupal.displace.calculateOffset('right');
      left = (left + width) < rightEdge ? left : left - ((left + width) - rightEdge);

      formItem.style.left = left + 'px';
      formItem.style.width = width + 'px';
      formItem.classList.add('form-item-popout');
    }
  }

  /**
   * Get remain focused selectors.
   *
   * When an input like a text area or an inline editor has focus, we don't always
   * want to hide the editor when the user clicks on something else on the page.
   * e.g. clicking a Bold icon in the CKEditor or changing a text format in the
   * sidebar.
   *
   * @returns {string[]}
   *   An array of selectors that will not cause the current form item to lose
   *   focus if clicked.
   */
  getFocusedSelectors() {
    return [
      '.edit-plus-focused',
      '#toolbar-plus-right-sidebar',
      '.ck-body-wrapper',
    ];
  }

  /**
   * Applies.
   *
   * When the plugin manager is selecting plugins, ensure that this plugin applies.
   *
   * @param EditableElement
   *   The editable element.
   * @returns {boolean}
   *   Whether this plugin should be used.
   */
  applies(EditableElement) {
    return true;
  }

}

